#!/usr/bin/env python3

from __future__ import annotations

import ipaddress
import json
import logging
import re
import socket

from importlib.metadata import version
from urllib.parse import urljoin, urlparse

import requests


class PySecurityTXTException(Exception):
    pass


class SecurityTXTNotAvailable(PySecurityTXTException):
    pass


class PySecurityTXT():

    def __init__(self, loglevel: int=logging.INFO, useragent: str | None=None,
                 *, proxies: dict[str, str] | None=None):
        """Do things to a security.txt file.

        :param loglevel: Python loglevel
        :param useragent: The User Agent used by requests to run the HTTP request.
        :param proxies: The proxies to use for the request. More details: https://requests.readthedocs.io/en/latest/
        """
        self.logger = logging.getLogger(f'{self.__class__.__name__}')
        self.logger.setLevel(loglevel)
        self.session = requests.session()
        self.session.headers['user-agent'] = useragent if useragent else f'PySecurityTXT / {version("pysecuritytxt")}'
        if proxies:
            self.session.proxies.update(proxies)
        self.expected_paths = ['/.well-known/security.txt', '/security.txt']

    def _try_get_url(self, url: str) -> str:
        try:
            response = self.session.get(url, timeout=4)
            response.raise_for_status()
        except requests.Timeout as e:
            raise e
        except (requests.HTTPError, requests.exceptions.ConnectionError):
            raise SecurityTXTNotAvailable(f'Unable to get the file from: {url}')
        return response.text

    def _all_possible_domains(self, hostname: str) -> set[str]:
        hostname_parts = hostname.split('.')
        if len(hostname_parts) <= 2:
            return {hostname, }
        current_domain = '.'.join(hostname_parts[-2:])
        to_return: set[str] = {current_domain, }
        for domain_part in reversed(hostname_parts[:-2]):
            current_domain = f'{domain_part}.{current_domain}'
            to_return.add(current_domain)
        return to_return

    def parse(self, file: str) -> dict[str, str | list[str]]:
        """Takes a security.txt file, parses it.

            :param file: The security.txt file.
        """
        to_return = {}
        if acknowledgments := re.findall("^[A,a]cknowledgments[:]? (.*)$", file, re.MULTILINE):
            if len(acknowledgments) == 1:
                to_return['acknowledgments'] = acknowledgments[0]
            else:
                to_return['acknowledgments'] = acknowledgments
        if canonical := re.findall("^[C,c]anonical[:]? (.*)$", file, re.MULTILINE):
            if len(canonical) == 1:
                to_return['canonical'] = canonical[0]
            else:
                to_return['canonical'] = canonical
        if contact := re.findall("^[C,c]ontact[:]? (.*)$", file, re.MULTILINE):
            if len(contact) == 1:
                to_return['contact'] = contact[0]
            else:
                to_return['contact'] = contact
        if csaf := re.findall("^CSAF[:]? (.*)$", file, re.MULTILINE):
            if len(csaf) == 1:
                to_return['csaf'] = csaf[0]
            else:
                to_return['csaf'] = csaf
        if encryption := re.findall("^[E,e]ncryption[:]? (.*)$", file, re.MULTILINE):
            if len(encryption) == 1:
                to_return['encryption'] = encryption[0]
            else:
                to_return['encryption'] = encryption
        if expires := re.findall("^[E,e]xpires[:]? (.*)$", file, re.MULTILINE):
            if len(expires) == 1:
                to_return['expires'] = expires[0]
            else:
                to_return['expires'] = expires
        if hiring := re.findall("^[H,h]iring[:]? (.*)$", file, re.MULTILINE):
            if len(hiring) == 1:
                to_return['hiring'] = hiring[0]
            else:
                to_return['hiring'] = hiring

        if policy := re.findall("^[P,p]olicy[:]? (.*)$", file, re.MULTILINE):
            if len(policy) == 1:
                to_return['policy'] = policy[0]
            else:
                to_return['policy'] = policy
        if prefered_languages := re.findall("^[P,p]referred-[L,l]anguages[:]? (.*)$", file, re.MULTILINE):
            # this one must be there only once, and contains a list of languages
            # I have limited trust on that, so let's normalize it
            languages: list[str] = []
            for lang_list in prefered_languages:
                languages += [language.strip() for language in lang_list.split(',') if language.strip()]
            to_return['prefered-languages'] = languages
        return to_return

    def get(self, hint: str, /, *, parse: bool=False) -> str:
        '''Get the security.txt file.

            :param hint: It can be a domain, an IP or an URL (and we try to figure out where the file is), or a full URL to the file.
            :param parse: Parse the file and returns a json dictionary with the relevant fields
        '''
        if hint.endswith('security.txt'):
            response = self._try_get_url(hint)
            if parse:
                return json.dumps(self.parse(response))
            return response

        if re.search("^http[s]?://", hint):
            # we have a URL, get the hostname
            splitted_url = urlparse(hint)
            if not (hostname := splitted_url.hostname):
                raise PySecurityTXTException(f'Unable to get a hostname out of the hint: {hint}')
        else:
            hostname = hint

        try:
            ipaddress.ip_address(hostname)
            all_domains = {hostname, }
            ip = True
        except ValueError:
            all_domains = self._all_possible_domains(hostname)
            ip = False
        test_urls: set[tuple[str, str]] = set()
        for domain in all_domains:
            # we have what should be a domain or an ip, let's try a few things
            try:
                if not ip:
                    # Quick domain name resolution and ignore if it fails
                    socket.gethostbyname(domain)
            except Exception as e:
                self.logger.info(f'Unable to resolve {domain}: {e}')
            else:
                for expected_path in self.expected_paths:
                    test_urls.update([
                        (domain, urljoin(f'https://{domain}', expected_path)),
                        (domain, urljoin(f'http://{domain}', expected_path))
                    ])

            if not domain.startswith('www') and not ip:
                www_domain = f'www.{domain}'
                try:
                    socket.gethostbyname(www_domain)
                except Exception as e:
                    self.logger.info(f'Unable to resolve {www_domain}: {e}')
                else:
                    for expected_path in self.expected_paths:
                        test_urls.update([
                            (www_domain, urljoin(f'https://{www_domain}', expected_path)),
                            (www_domain, urljoin(f'http://{www_domain}', expected_path))
                        ])

        timeout_domains = set()
        tested_urls = set()
        for domain, url in sorted(test_urls, key=len, reverse=True):  # Longest URL first, so /.well-known/ path is prioritized
            if domain in timeout_domains:
                continue
            tested_urls.add(url)
            try:
                response = self._try_get_url(url)
                if re.search("^[C,c]ontact", response, re.MULTILINE):
                    # Because we often get responses that are just plain HTML pages and a 2XX status code.
                    break
            except requests.Timeout:
                timeout_domains.add(domain)
            except SecurityTXTNotAvailable:
                self.logger.debug(f'Not available on {url}')
        else:
            raise SecurityTXTNotAvailable(f'Unable to file on {", ".join(tested_urls)}')

        if not parse:
            return response
        return json.dumps(self.parse(response))
